from sklearn.linear_model import LinearRegression
import numpy as np


class ProcesamientoRegresionLineal:
    def __init__(self,
                 nombre_regresion=None,
                 arreglo_a_procesar=None,
                 brinco_temporal=None,
                 longitud_ventana_tiempo=None):
        self.nombre_regresion = nombre_regresion
        self.arreglo_a_procesar = arreglo_a_procesar
        self.brinco_temporal = brinco_temporal
        self.longitud_ventana_tiempo = longitud_ventana_tiempo
        self.indices_tiempo = None
        self.series_tiempo = None
        self.resultados_regresion_lineal = None
        self.pendientes_regresion_lineal = None

    def realiza_procesamiento(self):
        self.indices_tiempo, self.series_tiempo = \
            generar_series_tiempo(lista_a_procesar=self.arreglo_a_procesar,
                                  brinco_temporal=self.brinco_temporal,
                                  longitud_ventana_tiempo=self.longitud_ventana_tiempo)

        self.resultados_regresion_lineal = \
            generar_regresion_lineal_para_series_tiempo(indices_tiempo=self.indices_tiempo,
                                                        series_tiempo=self.series_tiempo)

        self.pendientes_regresion_lineal = \
            genera_pendientes_de_rectas_regresion_lineal(self.indices_tiempo,
                                                         self.resultados_regresion_lineal)

    def regresa_resultados_regresion_lineal(self):
        return self.indices_tiempo,\
               self.series_tiempo,\
               self.resultados_regresion_lineal,\
               self.pendientes_regresion_lineal

    def genera_diccionario_informacion_actual(self):
        diccionario_informacion = dict()
        indices, series, resultados, pendientes = self.regresa_resultados_regresion_lineal()
        diccionario_informacion["indices_tiempo"] = indices
        diccionario_informacion["series_tiempo"] = series
        diccionario_informacion["resultados"] = resultados
        diccionario_informacion["pendientes"] = pendientes
        return diccionario_informacion

    def agrega_categoria_regresion_lineal(self,
                                          diccionario_regresiones_lineales,
                                          nombre_tipo_regresion_lineal):
        nuevo_diccionario = self.genera_diccionario_informacion_actual()
        diccionario_regresiones_lineales[nombre_tipo_regresion_lineal] = nuevo_diccionario


def generar_series_tiempo(lista_a_procesar, brinco_temporal, longitud_ventana_tiempo):
    series_tiempo_resultantes = []
    indices_tiempo_resultantes = []
    for brinco in range(0, len(lista_a_procesar), brinco_temporal):
        nueva_serie = []
        nueva_lista_indices = []
        if brinco+longitud_ventana_tiempo < len(lista_a_procesar):
            for indice_lista in range(brinco, brinco+longitud_ventana_tiempo, +1):
                nueva_serie.append(lista_a_procesar[indice_lista])
                nueva_lista_indices.append(indice_lista)
            series_tiempo_resultantes.append(nueva_serie)
            indices_tiempo_resultantes.append(nueva_lista_indices)
    return np.array(indices_tiempo_resultantes), np.array(series_tiempo_resultantes)


def generar_regresion_lineal_para_series_tiempo(indices_tiempo, series_tiempo):
    lista_predicciones = []
    regresor_lineal = LinearRegression()
    for indice in range(0, len(indices_tiempo)):
        X = indices_tiempo[indice].reshape(-1, 1)
        Y = series_tiempo[indice].reshape(-1, 1)
        regresor_lineal.fit(X, Y)
        prediccion = regresor_lineal.predict(X)
        lista_predicciones.append(prediccion)
    return lista_predicciones


def genera_pendientes_de_rectas_regresion_lineal(valores_x_regresion_lineal,
                                                 valores_y_regresion_lineal):
    pendientes = []
    cantidad_listas = len(valores_y_regresion_lineal)
    for indice in range(0, cantidad_listas):
        valores_x = valores_x_regresion_lineal[indice]
        valores_y = valores_y_regresion_lineal[indice]
        delta_y = valores_y[-1] - valores_y[0]
        delta_x = valores_x[-1] - valores_x[0]
        pendiente_actual = delta_y/delta_x
        pendientes.append(pendiente_actual)
    return pendientes






import numpy as np
import pandas as pd


def calcula_serie_velocidad(arreglo_tiempo,
                            posiciones_x, posiciones_y):
    tiempos = arreglo_tiempo
    longitud_movimientos = len(posiciones_x)
    lista_velocidad = np.zeros(longitud_movimientos)

    for i in range(1, longitud_movimientos):
        valor_tiempo = tiempos[i] - tiempos[i - 1]
        if np.isnan(posiciones_x.iloc[i - 1]):
            lista_velocidad[i - 1] == 0
        elif np.isnan(posiciones_x.iloc[i]):
            lista_velocidad[i - 1] == 0
        else:
            lista_velocidad[i - 1] = (((posiciones_x.iloc[i] - posiciones_x.iloc[i - 1]) ** 2 + (
                    posiciones_y.iloc[i] - posiciones_y.iloc[i - 1]) ** 2) ** 0.5) / valor_tiempo
    serie_velocidad = pd.Series(lista_velocidad)
    velocidad_maxima = round(serie_velocidad.max(), 3)

    return serie_velocidad


def calcula_serie_aceleracion(arreglo_tiempo,
                              serie_velocidad):
    tiempos = arreglo_tiempo
    longitud_velocidad = serie_velocidad.size
    lista_aceleracion = np.zeros(longitud_velocidad)

    for i in range(1, longitud_velocidad):
        valor_tiempo = tiempos[i] - tiempos[i - 1]
        lista_aceleracion[i - 1] =\
            (serie_velocidad.iloc[i] - serie_velocidad.iloc[i - 1]) / valor_tiempo

    serie_aceleracion = pd.Series(lista_aceleracion)
    aceleracion_maxima = round(serie_aceleracion.max(), 3)
    return serie_aceleracion, aceleracion_maxima


def obtener_arreglo_promediado(serie_a_promediar, cantidad_datos_promedio):
    longitud_serie_original = serie_a_promediar.size
    posicion_extra = 0
    if longitud_serie_original % cantidad_datos_promedio > 0:
        posicion_extra = 1
    longitud_arreglo_promediado =\
        int(longitud_serie_original / cantidad_datos_promedio + posicion_extra)
    arreglo_promediado = np.zeros(longitud_arreglo_promediado)
    for i in range(0, longitud_arreglo_promediado - 1):
        for j in range(cantidad_datos_promedio):
            arreglo_promediado[i] = \
                arreglo_promediado[i] + serie_a_promediar[cantidad_datos_promedio*i + j]
        arreglo_promediado[i] /= cantidad_datos_promedio
    for j in range(longitud_serie_original % cantidad_datos_promedio):
        arreglo_promediado[-1] = \
            serie_a_promediar.iloc[cantidad_datos_promedio*(longitud_arreglo_promediado-1) + j]
    serie_promediada = pd.Series(serie_a_promediar)
    return serie_promediada



class ConfigGrafica:
    def __init__(self):

        self.limite_eje_x_min = None
        self.limite_eje_x_max = None
        self.limite_eje_y_min = None
        self.limite_eje_y_max = None

        self.tomar_objeto_1 = True
        self.tomar_objeto_2 = True
        self.tomar_objeto_3 = True
        self.tomar_objeto_4 = True

        self.objeto_a_promediar = 1
        self.tamanio_intervalo_para_promedio = 5

        self.rot_grafica_3d = 220

        self.dist_max_figuras = 24
        self.total_figuras = 2
        self.tipo_figura = "Cilindros"

        self.columna_consecuencia_disponible = False

        self.lista_consecuencias = None
        self.consecuencia_elegida = None
        self.brinco_temporal_consecuencias = 1
        self.tamanio_vectores_consecuencias = 1

        self.tipos_regresion_lineal = ["Velocidad",
                                       "Aceleracion"]
        self.regresion_lineal_seleccionada = self.tipos_regresion_lineal[0]

        self.regiones_preferidas_3d = [0]

    def regresa_config_grafica_general(self):
        limites_eje_x = [self.limite_eje_x_min, self.limite_eje_x_max]
        limites_eje_y = [self.limite_eje_y_min, self.limite_eje_y_max]
        return limites_eje_x, limites_eje_y

    def regresa_config_grafica_distancias(self):
        objetos_a_utilizar = [self.tomar_objeto_1,
                              self.tomar_objeto_2,
                              self.tomar_objeto_3,
                              self.tomar_objeto_4]
        objeto_a_promediar = self.objeto_a_promediar
        tamanio_intervalo_para_promedio = self.tamanio_intervalo_para_promedio
        return objetos_a_utilizar, objeto_a_promediar, tamanio_intervalo_para_promedio

    def regresa_config_grafica_figuras(self):
        return self.dist_max_figuras, self.total_figuras, self.tipo_figura, self.rot_grafica_3d

    def regresa_config_grafica_regresion_lineal(self):
        return self.regresion_lineal_seleccionada

    def regresa_config_grafica_consecuencias(self):
        return self.brinco_temporal_consecuencias,\
               self.tamanio_vectores_consecuencias,\
               self.consecuencia_elegida



class InformacionRutasVectoriales:
    def __init__(self, posiciones_eje_x=None, posiciones_eje_y=None,
                 tamanio_vectorial=None, brinco_temporal=None,
                 indice_consecuencia=None):

        self.posiciones_eje_x = posiciones_eje_x
        self.posiciones_eje_y = posiciones_eje_y
        self.tamanio_vectorial = tamanio_vectorial
        self.brinco_temporal = brinco_temporal
        self.indice_consecuencia = indice_consecuencia

        self.posiciones_x_antes_consecuencia = None
        self.posiciones_y_antes_consecuencia = None

        self.posiciones_x_antes_consecuencia_con_consecuencia = None
        self.posiciones_y_antes_consecuencia_con_consecuencia = None

        self.posiciones_x_despues_consecuencia = None
        self.posiciones_y_despues_consecuencia = None

        self.direcciones_x_antes_consecuencia = None
        self.direcciones_y_antes_consecuencia = None
        self.direcciones_x_despues_consecuencia = None
        self.direcciones_y_despues_consecuencia = None

    def procesar_informacion_vectorial(self):
        self.obtener_posiciones_vectores()

        self.obtener_direcciones_vectores()


    def obtener_posiciones_vectores(self):
        self.posiciones_x_antes_consecuencia, self.posiciones_x_despues_consecuencia = \
            obtener_posiciones_correspondientes_direccion_eje(self.posiciones_eje_x,
                                                              self.indice_consecuencia,
                                                              self.tamanio_vectorial,
                                                              self.brinco_temporal)

        self.posiciones_y_antes_consecuencia, self.posiciones_y_despues_consecuencia = \
            obtener_posiciones_correspondientes_direccion_eje(self.posiciones_eje_y,
                                                              self.indice_consecuencia,
                                                              self.tamanio_vectorial,
                                                              self.brinco_temporal)

        self.posiciones_x_antes_consecuencia_con_consecuencia = self.posiciones_x_antes_consecuencia.copy()
        self.posiciones_x_antes_consecuencia_con_consecuencia.append(self.posiciones_eje_x[self.indice_consecuencia])

        self.posiciones_y_antes_consecuencia_con_consecuencia =  self.posiciones_y_antes_consecuencia.copy()
        self.posiciones_y_antes_consecuencia_con_consecuencia.append(self.posiciones_eje_y[self.indice_consecuencia])

    def obtener_direcciones_vectores(self):
        self.direcciones_x_antes_consecuencia, self.direcciones_y_antes_consecuencia = \
            obtener_vectores_posicion_flechas(self.posiciones_x_antes_consecuencia_con_consecuencia,
                                              self.posiciones_y_antes_consecuencia_con_consecuencia)

        self.direcciones_x_despues_consecuencia, self.direcciones_y_despues_consecuencia =\
            obtener_vectores_posicion_flechas(self.posiciones_x_despues_consecuencia,
                                              self.posiciones_y_despues_consecuencia)

    def regresa_posiciones_direcciones(self):
        arreglo_posiciones_x = [self.posiciones_x_antes_consecuencia,
                                self.posiciones_x_despues_consecuencia]

        arreglo_posiciones_y = [self.posiciones_y_antes_consecuencia,
                                self.posiciones_y_despues_consecuencia]

        arreglo_direcciones_x = [self.direcciones_x_antes_consecuencia,
                                 self.direcciones_x_despues_consecuencia]

        arreglo_direcciones_y = [self.direcciones_y_antes_consecuencia,
                                 self.direcciones_y_despues_consecuencia]

        return arreglo_posiciones_x, arreglo_posiciones_y, \
               arreglo_direcciones_x, arreglo_direcciones_y


def obtener_vectores_posicion_flechas(vector_x, vector_y):
    x_direct = []
    y_direct = []

    for i in range(0, len(vector_x) - 1):
        x1 = vector_x[i]
        x2 = vector_x[i + 1]
        y1 = vector_y[i]
        y2 = vector_y[i + 1]
        x_punta, y_punta = obtener_punta_vector(x1, x2, y1, y2)
        x_direct.append(x_punta)
        y_direct.append(y_punta)

    return x_direct, y_direct


def obtener_punta_vector(x_1, x_2, y_1, y_2):
    return x_2 - x_1, y_2 - y_1


def obtener_posiciones_correspondientes_direccion_eje(posiciones_eje_correspondiente,
                                                      consecuencia_elegida,
                                                      tamanio_vectores,
                                                      brinco_temporal):

    pos_antes_consecuencia = posiciones_eje_correspondiente.tolist()

    pos_antes_consecuencia = pos_antes_consecuencia[int(consecuencia_elegida - tamanio_vectores):
                                                    int(consecuencia_elegida):
                                                    int(brinco_temporal)]

    pos_despues_consecuencia = posiciones_eje_correspondiente.tolist()

    pos_despues_consecuencia = pos_despues_consecuencia[int(consecuencia_elegida):
                                                        int(consecuencia_elegida + tamanio_vectores):
                                                        int(brinco_temporal)]

    return pos_antes_consecuencia, pos_despues_consecuencia



from motus.experimento import Experimento
import motus.entropia_divergencia.entropia_divergencia as ed
from matplotlib.figure import Figure


class MultiplesExperimentos:

    def __init__(self,
                 lista_archivos,
                 dimension_caja=None,
                 cantidad_filas=10,
                 cantidad_columnas=10,
                 procesar_tiempo_archivo_irregular=False,
                 puntos_importantes=None,
                 objeto_a_promediar=1,
                 tamanio_intervalo_para_promedio=100,
                 arreglo_regiones_preferidas=""):
        self.lista_archivos = lista_archivos
        self.lista_experimentos = []
        self.resultados_entropia = {}
        self.resultados_divergencia = {}
        self.arreglo_regiones_preferidas = arreglo_regiones_preferidas

        if dimension_caja is None:
            self.dimension_caja = [100, 100]

        self.dimension_caja = dimension_caja
        self.cantidad_filas = cantidad_filas
        self.cantidad_columnas = cantidad_columnas
        self.procesar_tiempo_archivo_irregular = procesar_tiempo_archivo_irregular
        self.puntos_importantes = puntos_importantes
        self.objeto_a_promediar = objeto_a_promediar
        self.tamanio_intervalo_para_promedio = tamanio_intervalo_para_promedio

    def inicializa_experimentos(self):

        print("Procesar tiempo irregular: ", self.procesar_tiempo_archivo_irregular)
        if isinstance(self.lista_archivos, dict):

            count = 0

            print("Diccionario de archivos: ", self.lista_archivos)

            for (nombre, valor) in self.lista_archivos.items():
                if valor.get() is True:
                    count =+ 1
                    print("conteo de experimentos: ", count)
                    nuevo_experimento = Experimento(str(nombre),
                                                    dimension_caja=self.dimension_caja,
                                                    cantidad_filas=self.cantidad_filas,
                                                    cantidad_columnas=self.cantidad_columnas,
                                                    procesar_tiempo_archivo_irregular=
                                                    self.procesar_tiempo_archivo_irregular,
                                                    puntos_importantes=self.puntos_importantes)
                    nuevo_experimento.calcula_velocidades()
                    nuevo_experimento.calcula_regiones_cuadriculadas()
                    nuevo_experimento.calcula_matrices_solo_regiones_preferidas()
                    self.lista_experimentos.append(nuevo_experimento)
        else:
            for archivo in self.lista_archivos:
                nuevo_experimento = Experimento(archivo,
                                                dimension_caja=self.dimension_caja,
                                                cantidad_filas=self.cantidad_filas,
                                                cantidad_columnas=self.cantidad_columnas,
                                                procesar_tiempo_archivo_irregular=
                                                self.procesar_tiempo_archivo_irregular,
                                                puntos_importantes=self.puntos_importantes,
                                                )
                nuevo_experimento.calcula_velocidades()
                nuevo_experimento.calcula_regiones_cuadriculadas()
                nuevo_experimento.calcula_matrices_solo_regiones_preferidas()
                self.lista_experimentos.append(nuevo_experimento)
        self.procesa_diccionario_entropias()
        self.procesa_diccionario_divergencias()

    def procesa_diccionario_entropias(self):
        """
        Se lee uno de los diccionarios de entropía para conocer sus llaves. Por cada llave, se obtienen todos
        los datos correspondientes a esa entropía para cada experimento. La lista resultante se guarda en un
        diccionario, donde la llave es el tipo de entropía y el registro es la lista con los valores encontrados.

        El resultado es un diccionario con todas las entropías registradas, y listas con los valores de
        cada experimento
        """

        experimento_inicial = self.lista_experimentos[0]

        for tipo_entropia in sorted(experimento_inicial.entropias):
            valores_entropia = []
            for experimento in self.lista_experimentos:
                valor_entropia_experimento = experimento.entropias[tipo_entropia]
                valores_entropia.append(valor_entropia_experimento)
            self.resultados_entropia[tipo_entropia] = valores_entropia

    def procesa_diccionario_divergencias(self):
        lista_arreglos_regiones = list()
        lista_arreglos_solo_regiones_preferidas = list()
        lista_arreglos_solo_regiones_no_preferidas = list()
        for experimento in self.lista_experimentos:
            lista_arreglos_regiones.append(
                experimento.matriz_cuadriculada.flatten())
            lista_arreglos_solo_regiones_preferidas.append(
                experimento.matriz_cuadriculada_solo_regiones_preferidas.flatten())
            lista_arreglos_solo_regiones_no_preferidas.append(
                experimento.matriz_cuadriculada_solo_regiones_no_preferidas.flatten())
        self.calcula_divergencia("visitas_regiones", lista_arreglos_regiones)
        self.calcula_divergencia("visitas_regiones_preferidas", lista_arreglos_solo_regiones_preferidas)
        self.calcula_divergencia("visitas_regiones_no_preferidas", lista_arreglos_solo_regiones_no_preferidas)

    def calcula_divergencia(self, tipo_divergencia, lista_datos):
        lista_divergencia = ed.calcula_divergencias_de_lista_arreglos(lista_datos)
        self.resultados_divergencia[tipo_divergencia] = lista_divergencia

    def regresa_imagen_grafica(self, grafica):
        print("-----------------------------------------------")
        print("Gráfica deseada:", grafica)

        fig = Figure(figsize=(6, 6))
        fig_plot = fig.add_subplot(111)
        fig_plot.grid(True)

        cantidad_experimentos = len(self.lista_experimentos)

        if grafica is 'entropia_visitas_regiones':
            fig_plot.plot(range(1, cantidad_experimentos+1),
                          self.resultados_entropia["visitas_regiones"], 'o')
            fig_plot.plot(range(1, cantidad_experimentos+1),
                          self.resultados_entropia["visitas_regiones"])

        elif grafica is 'entropia_visitas_regiones_preferidas':
            fig_plot.plot(range(1, cantidad_experimentos+1),
                          self.resultados_entropia["visitas_regiones_preferidas"], 'o')
            fig_plot.plot(range(1, cantidad_experimentos+1),
                          self.resultados_entropia["visitas_regiones_preferidas"])

        elif grafica is 'entropia_visitas_regiones_no_preferidas':
            fig_plot.plot(range(1, cantidad_experimentos+1),
                          self.resultados_entropia["visitas_regiones_no_preferidas"], 'o')
            fig_plot.plot(range(1, cantidad_experimentos+1),
                          self.resultados_entropia["visitas_regiones_no_preferidas"])

        elif grafica is 'entropia_hist_velocidad':
            fig_plot.plot(range(1, cantidad_experimentos+1),
                          self.resultados_entropia["histograma_velocidad"], 'o')
            fig_plot.plot(range(1, cantidad_experimentos+1),
                          self.resultados_entropia["histograma_velocidad"])

        elif grafica is 'divergencia_regiones':
            fig_plot.plot(range(1, cantidad_experimentos),
                          self.resultados_divergencia["visitas_regiones"], 'o')
            fig_plot.plot(range(1, cantidad_experimentos),
                          self.resultados_divergencia["visitas_regiones"])

        elif grafica is 'divergencia_regiones_preferidas':
            fig_plot.plot(range(1, cantidad_experimentos),
                          self.resultados_divergencia["visitas_regiones_preferidas"], 'o')
            fig_plot.plot(range(1, cantidad_experimentos),
                          self.resultados_divergencia["visitas_regiones_preferidas"])

        elif grafica is 'divergencia_regiones_no_preferidas':
            fig_plot.plot(range(1, cantidad_experimentos),
                          self.resultados_divergencia["visitas_regiones_no_preferidas"], 'o')
            fig_plot.plot(range(1, cantidad_experimentos),
                          self.resultados_divergencia["visitas_regiones_no_preferidas"])

        return fig



from motus.regiones_cuadriculadas import regiones_cuadriculadas as m_reg_c
import numpy as np

def calcula_matrices_de_recurrencia(arreglo_regiones,
                                    region_importante=None):

    longitud_matriz_recurrencia = len(arreglo_regiones)

    matriz_recurrencia = np.zeros((longitud_matriz_recurrencia, longitud_matriz_recurrencia))

    if region_importante is None:

        for i in range(len(arreglo_regiones)):
            valor = arreglo_regiones[i]
            for j in range(len(arreglo_regiones)):
                if valor == arreglo_regiones[j]:
                    matriz_recurrencia[i][j] = 1

    elif type(region_importante) is int:
        for i in range(len(arreglo_regiones)):
            valor = arreglo_regiones[i]
            for j in range(len(arreglo_regiones)):
                if valor == arreglo_regiones[j]:
                    if valor == region_importante:
                        matriz_recurrencia[i][j] = 1

    elif type(region_importante) is list:

        indice_region_objeto_1 = region_importante[0]
        indice_region_objeto_2 = region_importante[1]
        indice_region_objeto_3 = region_importante[2]
        indice_region_objeto_4 = region_importante[3]

        for i in range(len(arreglo_regiones)):
            valor = arreglo_regiones[i]
            for j in range(len(arreglo_regiones)):
                if valor == arreglo_regiones[j]:
                    if valor == indice_region_objeto_1:
                        matriz_recurrencia[i][j] = 1
                    if valor == indice_region_objeto_2:
                        matriz_recurrencia[i][j] = 2
                    if valor == indice_region_objeto_3:
                        matriz_recurrencia[i][j] = 3
                    if valor == indice_region_objeto_4:
                        matriz_recurrencia[i][j] = 4

    return matriz_recurrencia